(module eval-tests mzscheme
  (require (lib "match.ss")
           "reify.scm"
           "test.scm")
  
  (define (to-mz exp)
    (cond
      [(and (pair? exp) (eq? (car exp) 'store))
       (cons 'let (rewrite-dots (cdr exp)))]
      [else (rewrite-dots exp)]))
  
  (define (rewrite-dots e)
    (let loop ([e e])
      (cond
        [(and (list? e)
              (memq 'dot e))
         (let l-loop ([e e])
           (cond
             [(eq? (cadr e) 'dot)
              (cons (loop (car e))
                    (loop (caddr e)))]
             [else (cons (loop (car e))
                         (l-loop (cdr e)))]))]
        [(list? e)
         (map loop e)]
        [else e])))
  
  (define (same-answer? prog expected)
    (match prog
      [`(store ,sfs ,got)
        (equal? got expected)]
      [else
       (error 'same-answer? "unknown prog ~s" prog)]))
  
  (define special-sharing-test
    '(store () ((lambda (f)
                  (eqv? (f) (eval (cons 'quote (cons (f) null)))))
                (lambda () '(x)))))
  
  (define (around-eval t)
    (parameterize ([current-reify reify/share])
      (t)))
  
  (define eval-ts
    (test-suite
     "eval.scm"
     to-mz
     same-answer?
     
     (test "procs"
           '(store () (- ((lambda (x) 2) 3)))
           -2)
     (test "procs"
           '(store () (- ((lambda (x) x) 3)))
           -3)
     (test "procs & set!"
           '(store () ((lambda (x) (begin (set! x 2) x)) 3))
           2)
     
     (test "set!"
           '(store () ((lambda (x) (begin (set! x (- x)) x)) 3))
           -3)
     
     (test "booleans"
           '(store () (if (car '(#f)) 1 2))
           '2)
     
     (test "booleans"
           '(store () (if (car '(#t)) 1 2))
           '1)
     
     (test "dotted pairs"
           '(store () (car '(1 dot 2)))
           '1)
     (test "dotted pairs"
           '(store () (cdr '(1 dot 2)))
           '2)
     
     (test "dotted pairs"
           '(store () (car (cdr '(0 1 2 x dot #f))))
           1)
     
     (test ""
           '(store () (car (car (cdr (cdr '((x) (y) (1)))))))
           '1)
     
     (test ""
           '(store () (eval '((lambda (x) x) 1)))
           '1)
     
     (test ""
           '(store () (eval (cons (lambda (x) x) (cons 1 null))))
           '1)
     
     (test ""
           '(store () ((lambda (f) (eval (cons f (cons 1 null)))) (lambda (x) x)))
           '1)
     (test ""
           '(store () (((lambda (y) (eval y)) '(lambda (x) x)) 1))
           '1)
     (test ""
           '(store () (car (eval '((lambda (x) x) '(1 2)))))
           '1)
     
     (test ""
           '(store () ((lambda (f) (eqv? (f) (f)))
                       (lambda () '(2))))
           '#t)
     
     (test ""
           '(store () (car (eval (quote (quote (1 dot 2))))))
           1)
     
     (test ""
           '(store () (cdr (eval (quote (quote (1 dot 2))))))
           2)
     
     (test ""
           '(store () '#t)
           '#t)
     
     (test ""
           '(store () (eval '(if #t #f #t)))
           '#f)
     
     (test ""
           '(store () (eqv? '(f) '(f)))
           '#f)
     
     (test ""
           '(store () ((lambda (f)
                         (eqv? (f) (eval (cons (cons 'lambda 
                                                     (cons null 
                                                           (cons (cons 'quote (cons (f) null)) null)))
                                               null))))
                       (lambda () '(x))))
           #f)
     
     (test "" special-sharing-test #f #t) ;; last #f means don't run MZ test
     
     ;; this test has to be last (because the r5 test suite assumes it is there and skips it)
     (test "" special-sharing-test #t #f around-eval)))
  
  
  (provide eval-ts))
