(module show-r5-examples mzscheme
  (require (planet "gui.ss" ("robby" "redex.plt" 2 4))
           (prefix r5rs: "r5rs.scm"))
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;;  functions that show terms, with and w/out proofs
  ;;
  
  (define (r5rs-show x) (traces/pred r5rs:lang r5rs:reductions (list x) no-proof-pred))
  (define (r5rs-no-proof-show x) (traces r5rs:lang r5rs:non-proof-reductions x))

  (provide r5rs-show r5rs-no-proof-show)
  
  (reduction-steps-cutoff 200)

  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;;  predicate that returns #t if a term
  ;;  imposes a proof obligation that two
  ;;  procedures are the same
  ;;
  
  (define (no-proof-pred sexp term-node)
    (let ([parents (term-node-parents term-node)])
      (when (null? parents)
        (definitely-no-proof term-node))
      (when (ormap (λ (parent) (hash-table-get definitely-no-proof-ht parent (λ () #f)))
                   (term-node-parents term-node))
        (unless (andmap (λ (parent-label) (regexp-match #rx"proof" parent-label))
                        (term-node-labels term-node))
          (definitely-no-proof term-node)))
      (unless (hash-table-get definitely-no-proof-ht term-node (λ () #f))
        (hash-table-put! currently-requires-proof term-node #t))
      (hash-table-get definitely-no-proof-ht term-node (λ () #f))))
  
  (define (definitely-no-proof term-node)
    (hash-table-put! definitely-no-proof-ht term-node #t)
    (when (hash-table-get currently-requires-proof term-node (λ () #f))
      (hash-table-remove! currently-requires-proof term-node)
      (for-each
       (λ (term-node)
         (let ([now-needs-proof? (needs-proof? term-node)])
           (unless now-needs-proof?
             (term-node-set-red! term-node #f)
             (hash-table-remove! currently-requires-proof term-node))))
       (hash-table-map currently-requires-proof (λ (term-node _) term-node)))))
  
  (define (needs-proof? term-node)
    (let ([parents (term-node-parents term-node)])
      (if (null? parents)
          #f
          (let ([parents-need-proof? (map needs-proof? parents)])
            (andmap (λ (parent-needs-proof? parent-label)
                      (or parent-needs-proof?
                          (and (not parent-needs-proof?)
                               (regexp-match #rx"proof" parent-label))))
                    parents-need-proof?
                    (term-node-labels term-node))))))
  
  (define definitely-no-proof-ht (make-hash-table))
  (define currently-requires-proof (make-hash-table))
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;; example uses of the above functions
  ;; #; comments out an entire sexpression.
  ;;

  #;
  (r5rs-show '(store () (dw () 
                            (define x 1) 
                            ((lambda (p q) x) 
                             (set! x (#%+ x 1))
                             (set! x (#%+ x 1))))))

  ;; the example from the commentary, with appropriate #%s.
  #;
  (r5rs-show '(store () (dw () 
                            (define values
                              (lambda (dot things)
                                (#%call/cc 
                                 (lambda (cont) (#%apply cont things)))))
                            (values 1 2 3))))
  
  ;; the example from the introduction to part ii, with appropriate #%s.
  #; 
  (r5rs-show '(store () (dw ()
                            (define x 1)
                            ((lambda (t)
                               (t)
                               (t))
                             (lambda ()
                               ((lambda (a b) x)
                                (set! x (#%+ x 1))
                                (set! x (#%* x 2)))))))))

